/*------------------------------------------------------------------------------
    GALE. Fine-grained parallel Genetic Algorithms for Data Mining.
    Copyright (C) 2002  Xavier Llor <xllora@illigal.ge.uiuc.edu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
------------------------------------------------------------------------------*/


package edu.uiuc.ge.illigal.gale.agents;

import edu.uiuc.ge.illigal.gale.data.Instance;
import edu.uiuc.ge.illigal.gale.data.AttributeSet;
import edu.uiuc.ge.illigal.gale.config.GALEConfiguration;

/** 
 * Defines a generic handle to Agents. It presents
 * the minimal set of functions that an agent (single or aggregated)
 * must provide.
 *
 * @author  Xavier Llor&agrave; &lt;<a href="mailto:xllora@illigal.ge.uiuc.edu">xllora@illigal.ge.uiuc.edu</a>&gt;
 * @version 0.9alpha 
 * @since   0.9alpha 
 *
 */

public interface Agent
{
	//-------------------------------------------------------------------------

	/**
	 * Provides the classification performance information. 
	 *
	 * @return The handle to the classification performance
	 */
	public AgentPerformance getPerformance ();

	//-------------------------------------------------------------------------

	/**
	 * Resets the agents' performance.
	 */
	public void resetPerformance ();

	//-------------------------------------------------------------------------

	/**
	 * Merges two agents.
	 *
	 * @param agn The second agent to be merged
	 * @return The merged agent
	 */
	public Agent merge ( Agent agn );

	//-------------------------------------------------------------------------

	/**
	 * Splits an agent
	 *
	 * @return The splitted agent
	 */
	public Agent split ();


	//-------------------------------------------------------------------------

	/**
	 * Clones the handled agent.
	 *
	 * @return The cloned agent
	 * @throws CloneNotSupportedException Thrown when the agent cannot be 
	 *                                    cloned
	 */
	public Object clone()
	throws CloneNotSupportedException;

	//-------------------------------------------------------------------------

	/**
	 * Returns a complexity measure.
	 *
	 * @return The complexity measure
	 */
	public float complexity ();

	//-------------------------------------------------------------------------

	/**
	 * Clean spurious issues.
	 */
	public Agent prune ();

	//-------------------------------------------------------------------------

	/**
	 * Classifies the given instance
	 *
	 * @param ex The instance to be classified
	 * @return The predicted class
	 */
	public int classify ( Instance ins );

	//-------------------------------------------------------------------------

	/**
	 * Returns the stringfied form of the agent
	 *
	 * @return The stringfied agent
	 */
	public String toString();

	//-------------------------------------------------------------------------

	/**
	 * Returns a random builded agent.
	 *
	 * @param galeCnf The GALE configuration object
	 * @param att The attributes set information object
	 * @return The agent
	 */
	public Agent buildAgent ( GALEConfiguration galeCnf, AttributeSet att );

	//-------------------------------------------------------------------------

}
