/*------------------------------------------------------------------------------
    GALE. Fine-grained parallel Genetic Algorithms for Data Mining.
    Copyright (C) 2002  Xavier Llor <xllora@illigal.ge.uiuc.edu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
------------------------------------------------------------------------------*/


package edu.uiuc.ge.illigal.gale.agents;

import edu.uiuc.ge.illigal.gale.config.*;
import edu.uiuc.ge.illigal.gale.data.*;
import edu.uiuc.ge.illigal.gale.distance.*;

import java.io.Serializable;
import java.util.Vector;
import java.util.Random;

/** 
 * Implements the generic handle to Agents for Instance Based Agents (GENIFER).
 *
 * @author  Xavier Llor&agrave; &lt;<a href="mailto:xllora@illigal.ge.uiuc.edu">xllora@illigal.ge.uiuc.edu</a>&gt;
 * @version 0.9alpha 
 * @since   0.9alpha 
 *
 */

public class IBAgent
implements Agent, Serializable, Cloneable
{
	//-------------------------------------------------------------------------

	/** The GALE environment configuration object */
   protected GALEConfiguration galeCnf;

	/** The attributes information for instances */
   protected AttributeSet attCnf;

   /** Agent Performance object*/
	protected AgentPerformance agnPer;

	/** Random number generator object */
	protected Random rnd;

	/** Distance function */
	protected Distance dis;

	/** Genetic material vector */
	protected Vector genVec;

	/** K number of instances */
	protected int iK;

	/** KNN minimal values */
	protected float [] fakNN;

	/** KNN minimal position */
	protected int [] iakNN;

	/** Prune information */
	protected int [] iaPrune;

	//------------------------------------------------------------------------

   /** 
	 * Builds a simple agent with null information
    */ 
      
   public IBAgent ()
   {
		attCnf  = null;
   	galeCnf = null;
		rnd     = null;
		dis     = null;
		agnPer  = null;
		genVec  = null;
		fakNN   = null;
		iakNN   = null;
		iaPrune = null;
   }
   
	//------------------------------------------------------------------------

   /** 
	 * Builds a simple agent. The Agent is build using the GALE configuration
    * object.
	 *
    * @param gale The GALE configuration
	 * @param att The attributes set information object
    */ 
      
   public IBAgent ( GALEConfiguration gale, AttributeSet att )
   {
		attCnf  = att;
   	galeCnf = gale;
		rnd     = gale.getRandom();
		dis     = gale.getDistance();
		agnPer  = new BatchPerformance(galeCnf);

		genVec  = new Vector();

		// Genetic Material initialization
		float [] faTmp = null;
		int iAtts = gale.getNumAttributes();
		for ( int i=0, iMax=gale.getAgentMaxInitWidth() ; i<iMax ; i++ ) {
			faTmp = new float[iAtts+1];
			for ( int j=0 ; j<iAtts ; j++ )
				faTmp[j] = boundedFloat ( att.attribute(j).lowerBound(),
				                          att.attribute(j).upperBound() );
			faTmp[iAtts] = rnd.nextInt(gale.getNumClasses());
			genVec.addElement(faTmp);
		}
		
		// Number of instance for kNN
		iK = gale.getAgentMaxInitDepth();

		fakNN = new float[iK];
		iakNN = new int[iK];

		// Purge information
		iaPrune = new int[genVec.size()];
   }
   
	//---------------------------------------------------------------------
	
	// Agent implementation
   public Agent buildAgent ( GALEConfiguration galeCnf, 
	                                    AttributeSet att )
	{
		return new IBAgent ( galeCnf, att );
	}

	//-------------------------------------------------------------------------

	// Agent implementation
	public void resetPerformance ()
	{
		agnPer.reset();
		for ( int i=0 ; i<iaPrune.length ; i++ )
			iaPrune[i] = 0;
	}

	//-------------------------------------------------------------------------

	// Agent implementation
	public AgentPerformance getPerformance ()
	{
		return agnPer;
	}

	//-------------------------------------------------------------------------

	// Agent implementation
	public Agent merge ( Agent agn )
	{
		IBAgent agnRes = null;
		IBAgent  agnIn = (IBAgent)agn;
		float []    fa = null;
		float    faTmp = 0;

		// Cut points for attributes
		int iAtts = galeCnf.getNumAttributes();
		int iCAP1 = rnd.nextInt(iAtts+1);
		int iCAP2 = rnd.nextInt(iAtts+1);

		// Computing the size of the agents
		int iST = genVec.size();
		int iSA = agnIn.genVec.size();

		// Cut points for the instances of this agent
		int iCITP1 = rnd.nextInt(iST);
		int iCITP2 = rnd.nextInt(iST);
		if ( iCITP2<iCITP1 ) {
			int iTmp = iCITP1;
			iCITP1   = iCITP2;
			iCITP2   = iTmp;
		}

		// Cut points for the instances of the input agent
		int iCIAP1 = rnd.nextInt(iSA);
		int iCIAP2 = rnd.nextInt(iSA);
		if ( iCIAP2<iCIAP1 ) {
			int iTmp = iCIAP1;
			iCIAP1   = iCIAP2;
			iCIAP2   = iTmp;
		}

		// Check if attribute cut points must be swapped
		if ( iCITP1==iCITP2 || iCIAP1==iCIAP2 ) 
			if ( iCAP2<iCAP1 ) {
				int iTmp = iCAP1;
				iCAP1    = iCAP2;
				iCAP2    = iTmp;
			}

		// Prepare the crossover structures
		int    iGenLen = iAtts+1;
		int iMaxLength = iST*iGenLen+iSA*iGenLen;
		float [] faT   = new float[iST*iGenLen];
		float [] faA   = new float[iSA*iGenLen];
		float [] faRT  = new float[iMaxLength];
		float [] faRA  = new float[iMaxLength];

		for ( int iArr=0, i=0 ; i<iST ; i++ ) {
			fa = (float[])genVec.elementAt(i);
			for ( int j=0 ; j<iGenLen ; j++, iArr++ )
				faT[iArr] = fa[j];
		}

		for ( int iArr=0, i=0 ; i<iSA ; i++ ) {
			fa = (float[])agnIn.genVec.elementAt(i);
			for ( int j=0 ; j<iGenLen ; j++, iArr++ )
				faA[iArr] = fa[j];
		}

		// Exchange the material
		int iXT1 = iCITP1*iGenLen+iCAP1;
		int iXT2 = iCITP2*iGenLen+iCAP2;
		int iXA1 = iCIAP1*iGenLen+iCAP1;
		int iXA2 = iCIAP2*iGenLen+iCAP2;

		int iT  = 0;
		int iA  = 0;
		int iRT = 0;
		int iRA = 0;

		// Coping the head of the genes
		for ( ; iT<iXT1 ; iT++, iRT++ ) faRT[iRT] = faT[iT]; 	
		for ( ; iA<iXA1 ; iA++, iRA++ ) faRA[iRA] = faA[iA]; 	

		// Swapping genes
		for ( ; iA<iXA2 ; iA++, iRT++ ) faRT[iRT] = faA[iA];
		for ( ; iT<iXT2 ; iT++, iRA++ ) faRA[iRA] = faT[iT];
		
		// Coping the tail of the genes
		for ( ; iT<iST*iGenLen ; iT++, iRT++ ) faRT[iRT] = faT[iT]; 	
		for ( ; iA<iSA*iGenLen ; iA++, iRA++ ) faRA[iRA] = faA[iA]; 	
		

		// Interpreting the resultant genetic material
		Vector vec = null;

		if ( iRT== 0 )
			vec = translateArray(faRA,iRA);
		else if ( iRA== 0 )
			vec = translateArray(faRT,iRT);
		else
			vec = (rnd.nextBoolean()) ? translateArray(faRT,iRT)
		   	                       : translateArray(faRA,iRA);
		
		try {
			agnRes         = (IBAgent)clone();
			agnRes.genVec  = vec;
			agnRes.iaPrune = new int[vec.size()];
		}
		catch ( Exception e ) {
			System.err.println("Fail in merge for "+getClass().getName());
			e.printStackTrace();
			System.exit(1);
		}

		return agnRes;
	}

	//-------------------------------------------------------------------------

	// Agent implementation
	public Agent split ()
	{
		IBAgent agn = null;

		try{
			agn = (IBAgent)clone();
			agn.mutateGenotype();
		}
		catch ( Exception e ) {
			System.err.println("Fail in split for "+getClass().getName());
			e.printStackTrace();
			System.exit(1);
		}

		return agn;
	}

	//-------------------------------------------------------------------------

	// Agent implementation
	public Object clone()
	throws CloneNotSupportedException
	{
		IBAgent nag = (IBAgent)super.clone();

      nag.galeCnf = galeCnf;
		nag.attCnf  = attCnf;
		nag.rnd     = rnd;
		nag.dis     = dis;
		nag.agnPer  = (AgentPerformance)agnPer.clone();

		nag.genVec = new Vector();
		for ( int i=0, iMax=genVec.size() ; i<iMax ; i++ )
			nag.genVec.addElement(((float[])genVec.elementAt(i)).clone());

		nag.iK = iK;
		nag.fakNN = (float[])fakNN.clone();
		nag.iakNN = (int[])iakNN.clone();

		nag.iaPrune = (int[])iaPrune.clone();

      return nag;
	}

	//-------------------------------------------------------------------------

	// Agent implementation
	public float complexity ()
	{
		return genVec.size();
	}

	//-------------------------------------------------------------------------

	// Agent implementation
	public Agent prune ()
	{
		int    iVec  = 0;
		int    iArr  = 0;
		int    iMax  = genVec.size();
		int [] iaTmp = null;

		for ( ; iArr<iMax ; iArr++ ) {
			if ( iaPrune[iArr]==0 ) {
				genVec.removeElementAt(iVec);
			}
			else 
				iVec++;
		}

		if ( genVec.size()==0 ) 
			return null;
		else if ( iMax==genVec.size() )
			return this;
		else {
			iArr = iVec = 0;
			iaTmp = new int[genVec.size()];
			while ( iArr<iMax ) {
				if ( iaPrune[iArr]!=0 ) {
					iaTmp[iVec] = iaPrune[iArr];
					iArr++;
					iVec++;
				}
				else
					iArr++;
			}
			iaPrune = iaTmp;
			return this;
		}
	}

	//-------------------------------------------------------------------------

	// Agent implementation
	public int classify ( Instance ins )
	{
		int   iMax    = genVec.size();
		float fTmp    = -1;
		float [] fa   = null;
		int   iCls    = -1;
		int   i       = 0;
		int   j       = 0;
		int   iFree   = iK;
		boolean bFlag = true;
		int   iNC     = galeCnf.getNumClasses();
		int   iNA     = galeCnf.getNumAttributes();
		int [] iaVot  = new int[iNC+1];
	
		// Resets the kNN information
		for ( ; j<iK ; j++ ) {
			fakNN[j]  = 1000*dis.maximalDistance();
			iakNN[j]  = -1;
		}

		// Computes the minimal k distances
		for ( i=0 ; i<iMax ; i++ ) {
			fa   = (float[])genVec.elementAt(i);
			fTmp = dis.distance(fa,0,ins,ins.getLength());
			for ( j=0, bFlag=true ; bFlag && j<iK ; j++ ) 
				if ( iakNN[j]==-1 ) {
					fakNN[j] = fTmp;
					iakNN[j] = i;
					bFlag    = false;
					iFree--;
				}
				else if ( fTmp<fakNN[j] && iFree==0 ) {
					fakNN[j] = fTmp;
					iakNN[j] = i;
					bFlag    = false;
				}
		}

		// Voting for the k nearest neighbors
		for ( i=0 ; i<iK && iakNN[i]!=-1 ; i++ ) {
			fa = (float[])genVec.elementAt(iakNN[i]);
			iaVot[(int)fa[iNA]+1]++;
			iaPrune[iakNN[i]]++;
		}

		// Choose the most voted class
		for ( i=0, j=-1, iCls=-1 ; i<iNC+1 ; i++ ) {
			if ( iaVot[i]>j ) {
				j = iaVot[i];
				iCls = i;
			} 
		}
		iCls--;	

		// Updates the classification performances
		agnPer.update(iCls,ins.getInstanceClass());

		return iCls;
	}

	//-------------------------------------------------------------------------

	// Agent implementation
	public String toString()
	{
		String sRes = "Instances of the model\n----------------------\n\n";
		float [] fa = null;
		int    iAtt = galeCnf.getNumAttributes();

		sRes += "Number of instances : "+genVec.size()+"\n\n";

		for ( int i=0, iMax=genVec.size() ; i<iMax ; i++ ) {
			fa = (float[])genVec.elementAt(i);
			for ( int j=0 ; j<iAtt ; j++ ) {
				sRes += fa[j]+"\t";
			}
			sRes += ((int)fa[iAtt])+"("+iaPrune[i]+")\n";
		}

		sRes += "\n"+agnPer;
			
		return sRes;
	}

	//-------------------------------------------------------------------------

	/**
    * Generates a random bounded float using the random number generator.
	 *
	 * @param lb lower bound
	 * @param ub upper bound
	 */
	protected float boundedFloat ( float lb, float ub )
	{
		float fDiff = ub-lb;

		fDiff = rnd.nextFloat()*fDiff;

		return fDiff+lb;
	}
	
	//-------------------------------------------------------------------------

	/**
    * Mutates the genetic material of instances.
	 */
	protected void mutateGenotype ()
	{
		int      iAtts = galeCnf.getNumAttributes();
		int       iCls = galeCnf.getNumClasses();
		float      fPM = galeCnf.getSomaticMP();
		float [] faTmp = null;

		for ( int i=0, iMax=genVec.size() ; i<iMax ; i++ ) {
			faTmp = (float[])genVec.elementAt(i);
			for ( int j=0 ; j<iAtts ; j++ )
				if ( rnd.nextFloat()<fPM )
					faTmp[j] = boundedFloat ( attCnf.attribute(j).lowerBound(),
				  	                          attCnf.attribute(j).upperBound() );
			if ( rnd.nextFloat()<fPM )
				faTmp[iAtts] = rnd.nextInt(iCls);
		}
	}

	//-------------------------------------------------------------------------

	/**
    * Build the vector representation of a genetic array.
	 *
	 * @param faGen the float array to translate
	 * @param iLen the length of useful information in the array
	 * @return the vector
	 */
	protected Vector translateArray ( float [] faGen, int iLen )
	{
		int   iAtts = galeCnf.getNumAttributes();
		int   iMaxs = iAtts+1;
		int   iRows = iLen/iMaxs;
		Vector  vec = new Vector();
		float [] fa = null;
		int    iAcc = 0;

		if ( iLen%iMaxs != 0 ) {
			System.err.println("Mistaken size after crossover "+iLen+","+iMaxs);
			System.exit(1);
		}
		
		for ( int i=0 ; i<iRows ; i++ ) {
			fa = new float[iMaxs];
			for ( int j=0 ; j<iMaxs ; j++, iAcc++ ) {
				fa[j] = faGen[iAcc];
			}
			vec.addElement(fa);
		}

		return vec;
	}

	//-------------------------------------------------------------------------

	public static void main ( String [] sArgs )
	{
		try {
			GALEConfiguration gg = new FileGALEConfiguration("iris.cnf");
			InstanceSet       ii = new FileInstanceSet(gg,"iris.arff");

			gg.getDistance().init(ii.attributeSet());

			Agent a = new IBAgent(gg,ii.attributeSet());
			//Agent b = new IBAgent(gg,ii.attributeSet());

			for ( int i=0, iMax=ii.getSize() ; i<iMax ; i++ ) {
				a.classify(ii.getInstance(i));	
				//b.classify(ii.getInstance(i));	
			}

			System.out.println(a);
			a = a.prune();
			//b = b.prune();
			System.out.println(a);
			//System.out.println(b);
			//Agent c = a.merge(b);
			//System.out.println(c);
		}
		catch ( Exception e ) {
			e.printStackTrace();
		}
	}

}
