/*------------------------------------------------------------------------------
    GALE. Fine-grained parallel Genetic Algorithms for Data Mining.
    Copyright (C) 2002  Xavier Llor <xllora@illigal.ge.uiuc.edu>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
------------------------------------------------------------------------------*/


package edu.uiuc.ge.illigal.gale.board;

import java.util.Random;
import java.util.Vector;
import java.io.*;

import edu.uiuc.ge.illigal.gale.data.*;
import edu.uiuc.ge.illigal.gale.config.*;
import edu.uiuc.ge.illigal.gale.agents.*;
import edu.uiuc.ge.illigal.gale.io.*;

/**
 * This class contains the Ale proposal based on CGA
 *
 * @author  Xavier Llor&agrave; &lt;<a href="mailto:xllora@illigal.ge.uiuc.edu">xllora@illigal.ge.uiuc.edu</a>&gt;
 * @version 0.9alpha
 * @since 0.9alpha
 */


public class Board
{
	//-----------------------------------------------------------------------

	/** GALE configuration object */
	protected GALEConfiguration galeCnf;

	/** GALE training instance set */
	protected InstanceSet insSet;

	/** The array of fold to use for cross-validation */
	protected OrderedSet [] osaFolds;

	/** IOFactory used for logs */
	protected IOFactory ioFac;

	/** Random number generator */
	protected Random rnd;

	//-----------------------------------------------------------------------

	/** The 2D board */
   protected Agent [][][] agnGen;

	/** The current board in use */
	protected int iBoard;

	/** The current agent performance */
	protected AgentPerformance [][] agnPer;

	/** Maximum width */
	protected int iMaxX;

	/** Maximum height */
	protected int iMaxY;

	/** The mapping of the folder in the training set */
	protected OrderedSet [][] osaMapping;

	/** The mapping of the folder in the training set contains all the 
	    instances */
	protected boolean [][] baAllIns;

	/** Best Agent found along the run */
	protected Agent agnBest;

	/** Dump board flag */
	protected boolean bDump;

	//-----------------------------------------------------------------------

	/**
    * Builds a board initializing all the data.
	 *
	 * @param gale The GALE environment configuration object
	 * @param ins The instance set available
	 * @param osa The array of folds for cross-validation
	 * @param io The IOFactory to use
	 * @param bD Dumps the board trace
    */
	public Board ( GALEConfiguration gale, InstanceSet ins, OrderedSet [] osa,
						IOFactory io, boolean bD )
	{
		// Updating references
		galeCnf  = gale;
		insSet   = ins;
		osaFolds = osa;
		rnd      = gale.getRandom();
		ioFac 	= io;
		bDump 	= bD;

		// Board arrangements
		iMaxX  = gale.getMaxX();
		iMaxY  = gale.getMaxY();
		agnGen = new Agent[2][iMaxY][iMaxX];
		agnPer = new AgentPerformance[iMaxY][iMaxX];
		iBoard = 0;
	}

	//-----------------------------------------------------------------------

	/**
    * Builds a board initializing all the data.
	 *
	 * @param iFold the fold to be run. -1 runs the whole instance set.
	 * @param sPre the preface file information
	 * @throws Exception Not valid mapping strategy
    */
	public void runFold ( int iFold, String sPre )
	throws Exception
	{	
		// The fold ID
		String sFold = "";

		if (iFold==-1) sFold="model";
		else {
			sFold += iFold;
			for ( int i=6, j=sFold.length() ; i>j ; i-- )
				sFold = "0"+sFold;
			sFold = "f"+sFold;
		}
		// Initializing the channels	
		ioFac.openChannel("traceEval","GALE."+sPre+".board."+sFold+".summary.accuracy"); 
		ioFac.openChannel("traceCmplx","GALE."+sPre+".board."+sFold+".summary.complexity"); 
		if ( bDump ) {
			ioFac.openChannel("eval","GALE."+sPre+".board."+sFold+".accuracy"); 
			ioFac.openChannel("cmplx","GALE."+sPre+".board."+sFold+".complexity"); 
		}	

		//  Variables
		int iInit  = 0;
		int iDepth = 0;
		int iX = 0;
		int iY = 0;
		OrderedSet osAll = null;
	
		// Build the examples to use
		if (iFold==-1)	osAll = insSet;
		else           osAll = new CVView(osaFolds,iFold);

		// Cleaning the board
		for (int i=0 ; i<iMaxY ; i++ ) 
   		for (int  j=0 ; j<iMaxX ; j++ ) {
				agnGen[iBoard][i][j] = null;
				agnPer[i][j]         = null;
			}

		// Initializes the agents
	
		iInit = (int)(galeCnf.getInitialAggregates()*iMaxX*iMaxY);
		for ( int i=0 ; i<iInit ; i++ ) {
	     	iY = rnd.nextInt(iMaxY);
	     	iX = rnd.nextInt(iMaxX);
         if ( agnGen[iBoard][iY][iX]==null ) {
        		agnGen[iBoard][iY][iX] = 
				           galeCnf.getAgent().buildAgent(galeCnf,insSet.attributeSet());
				agnPer[iY][iX] = agnGen[iBoard][iY][iX].getPerformance();
		   }
         else 
				i--;
		}

		// Initializes the best agent found
		agnBest = (Agent)agnGen[iBoard][iY][iX].clone();
       
		// Prepares the mapping information
		osaMapping = new OrderedSet[iMaxY][iMaxX];
		baAllIns   = new boolean[iMaxY][iMaxX];
		arrangeMapping(osAll);

		// Computes the performance of agents
      computePerformance();

		// Evolve the board
		initializeStatistics();
		evolveBoard();

		// Closing the channels	
		ioFac.closeChannel("traceEval");
		ioFac.closeChannel("traceCmplx");
		if ( bDump ) {
			ioFac.closeChannel("eval");
			ioFac.closeChannel("cmplx");
		}
	}

	//---------------------------------------------------------------------
    
	/** 
	 * Returns the best agent obtained along the run.
	 *
	 * @return the best agent
	 */
	public Agent bestAgent()
	{
		return agnBest;
	}

	//---------------------------------------------------------------------
    
	/** 
	 * Returns the full mapped agents
	 *
	 * @return the set of agents
	 */
	public Agent [] fullMappedAgents()
	{
		int i = 0;
		int j = 0;
		int iMax = 0;
		Vector vec  = new Vector();
		Agent [] aa = null;

   	for ( i=0 ; i<iMaxY ; i++ ) 
   		for ( j=0 ; j<iMaxX ; j++ ) 
				if ( baAllIns[i][j] && agnGen[iBoard][i][j]!=null )
					vec.addElement(agnGen[iBoard][i][j]);

		iMax=vec.size();
		aa = new Agent[iMax];
		for ( i=0, iMax=vec.size() ; i<iMax ; i++ )
			aa[i] = (Agent)vec.elementAt(i);
	
		return aa;
	}


	//---------------------------------------------------------------------
    
	/** 
	 * Evolve the board
	 * 
	 * @throws Exception exception raise in some of the phase of GALE
	 */
	protected void evolveBoard()
	throws Exception
	{
		int iIter = 0;
		int iMax  = galeCnf.getWorldIterations();

		for ( ; iIter<iMax ; iIter++ ) {
			computeStatistics(iIter,iMax);	
			merge();
			split();
      	computePerformance();
			survival();
		}
		computeStatistics(iIter,iMax);
	}

	//---------------------------------------------------------------------

	/** 
	 * Split phase.
	 * 
	 * @throws Exception exception raise in some of the phase of GALE
	 */
	public void split ()
	throws Exception
	{
		int [] iaSP       = null;
		Agent  agnNew     = null;
		float  fThreshold = 0.0f;

		for (int i=0 ; i<iMaxY ; i++ ) 
   		for (int  j=0 ; j<iMaxX ; j++ ) 
				if ( agnGen[iBoard][i][j]!=null ) {
					fThreshold = agnPer[i][j].getScaledAccuracy()*
					             galeCnf.getMaxSplitP(); 
				   if ( rnd.nextFloat()<fThreshold ) {
						iaSP = splitTargetPosition(i,j);
 						agnNew = agnGen[iBoard][i][j].split();
						agnGen[iBoard][iaSP[0]][iaSP[1]] = agnNew;
						agnPer[iaSP[0]][iaSP[1]] = agnNew.getPerformance();
					}
				}
	}

	//---------------------------------------------------------------------

	/** 
	 * Calculates the destination of the splitted agent on the board.
	 *	
	 * @param i The Y position of the agent
	 * @param j The X position of the agent
  	 * @return The target position 
	 */
	protected int [] splitTargetPosition( int i, int j )
	{
		int [] iaPos = new int[2];
		int    iNei  = -1;
		int    iTmp;

		int iUp    = (i+1)%iMaxY;
      int iDown  = (i+iMaxY-1)%iMaxY;
      int iRight = (j+1)%iMaxX;
      int iLeft  = (j+iMaxX-1)%iMaxX;
        
      iaPos[0] = -1; 
      iaPos[1] = -1; 

		iTmp = countNeighbours(iUp,iLeft);
      if ( agnGen[iBoard][iUp][iLeft]==null && iTmp>iNei ) {
			iaPos[0] = iUp;
			iaPos[1] = iLeft;
			iNei     = iTmp;
		}
        
		iTmp = countNeighbours(iUp,j);
      if ( agnGen[iBoard][iUp][j]==null && iTmp>iNei ) {
			iaPos[0] = iUp;
			iaPos[1] = j;
			iNei     = iTmp;
		}
        
		iTmp = countNeighbours(iUp,iRight);
      if ( agnGen[iBoard][iUp][iRight]==null && iTmp>iNei ) {
			iaPos[0] = iUp;
			iaPos[1] = iRight;
			iNei     = iTmp;
		}
        
		iTmp = countNeighbours(i,iLeft);
      if ( agnGen[iBoard][i][iLeft]==null && iTmp>iNei ) {
			iaPos[0] = i;
			iaPos[1] = iLeft;
			iNei     = iTmp;
		}
        
		iTmp = countNeighbours(i,iRight);
      if ( agnGen[iBoard][i][iRight]==null && iTmp>iNei ) {
			iaPos[0] = i;
			iaPos[1] = iRight;
			iNei     = iTmp;
		}
        
		iTmp = countNeighbours(iDown,iLeft);
      if ( agnGen[iBoard][iDown][iLeft]==null && iTmp>iNei ) {
			iaPos[0] = iDown;
			iaPos[1] = iLeft;
			iNei     = iTmp;
		}
        
		iTmp = countNeighbours(iDown,j);
      if ( agnGen[iBoard][iDown][j]==null && iTmp>iNei ) {
			iaPos[0] = iDown;
			iaPos[1] = j;
			iNei     = iTmp;
		}
        
		iTmp = countNeighbours(iDown,iRight);
      if ( agnGen[iBoard][iDown][iRight]==null && iTmp>iNei ) {
			iaPos[0] = iDown;
			iaPos[1] = iRight;
			iNei     = iTmp;
		}

		if ( iNei>-1)
			return iaPos;
		else {
			//No free space available
			return worstNeighbour(i,j);
		}
	}

	//---------------------------------------------------------------------

	/** 
	 * Returns the worst neighbor for the given cell.
	 *	
	 * @param i The Y position of the agent
	 * @param j The X position of the agent
  	 * @return The worst position 
	 */
   private int [] worstNeighbour( int i, int j )
   {
        int iUp    = (i+1)%iMaxY;
        int iDown  = (i+iMaxY-1)%iMaxY;
        int iRight = (j+1)%iMaxX;
        int iLeft  = (j+iMaxX-1)%iMaxX;
        
        float fBest  = 1000;
        
        int [] iaRes = new int[2];                
        iaRes[0] = iUp;
        iaRes[1] = iLeft;
        
        if ( agnGen[iBoard][iUp][iLeft]!=null )
			 if ( fBest>agnPer[iUp][iLeft].getScaledAccuracy() ) {
            fBest = agnPer[iUp][iLeft].getScaledAccuracy();
            iaRes[0] = iUp;
            iaRes[1] = iLeft;
        	 }
        if ( agnGen[iBoard][iUp][j]!=null )
          if ( fBest>agnPer[iUp][j].getScaledAccuracy() ) {
            fBest = agnPer[iUp][j].getScaledAccuracy();
            iaRes[0] = iUp;
            iaRes[1] = j;
        }
        if ( agnGen[iBoard][iUp][iRight]!=null )
          if ( fBest>agnPer[iUp][iRight].getScaledAccuracy() ) {
            fBest = agnPer[iUp][iRight].getScaledAccuracy();
            iaRes[0] = iUp;
            iaRes[1] = iRight;
        }
        
        if ( agnGen[iBoard][i][iLeft]!=null )
          if ( fBest>agnPer[i][iLeft].getScaledAccuracy() ) {
            fBest = agnPer[i][iLeft].getScaledAccuracy();
            iaRes[0] = i;
            iaRes[1] = iLeft;
        }
        if ( agnGen[iBoard][i][iRight]!=null )
          if ( fBest>agnPer[i][iRight].getScaledAccuracy() ) {
            fBest = agnPer[i][iRight].getScaledAccuracy();
            iaRes[0] = i;
            iaRes[1] = iRight;
        }
        if ( agnGen[iBoard][iDown][iLeft]!=null )
          if ( fBest>agnPer[iDown][iLeft].getScaledAccuracy() ) {
            fBest = agnPer[iDown][iLeft].getScaledAccuracy();
            iaRes[0] = iDown;
            iaRes[1] = iLeft;
        }
        if ( agnGen[iBoard][iDown][j]!=null )
          if ( fBest>agnPer[iDown][j].getScaledAccuracy() ) {
            fBest = agnPer[iDown][j].getScaledAccuracy();
            iaRes[0] = iDown;
            iaRes[1] = j;
        }
        if ( agnGen[iBoard][iDown][iRight]!=null )
          if ( fBest>agnPer[iDown][iRight].getScaledAccuracy() ) {
            fBest = agnPer[iDown][iRight].getScaledAccuracy();
            iaRes[0] = iDown;
            iaRes[1] = iRight;
        }

			
        return iaRes;
    }

	//---------------------------------------------------------------------
    
	/** 
	 * Implements the merge phase
	 * 
	 * @throws Exception could not clone an agent
	 */
	protected void merge()
	throws Exception
	{
		int   i, j;
		int   iaPos[] = new int[2];

		int iNextBoard = (iBoard+1)%2;

   	for ( i=0 ; i<iMaxY ; i++ ) 
   		for ( j=0 ; j<iMaxX ; j++ ) 
				if ( agnGen[iBoard][i][j]!=null ) {
					Agent agn = crossingNeighbour(i, j);
					if ( agn!=null && rnd.nextFloat()<galeCnf.getMergeP() ) {
						Agent agnNew = agnGen[iBoard][i][j].merge(agn);
						agnGen[iNextBoard][i][j] = agnNew;
					}
					else
						agnGen[iNextBoard][i][j] = agnGen[iBoard][i][j];
				}
				else
					agnGen[iNextBoard][i][j] = null;

		// Switching the board
		iBoard = iNextBoard;

		// Update performance references
		updatePerformance();
	}

	//---------------------------------------------------------------------

	/** 
	 * Returns the crossingNeighbour.
	 *
	 * @param i The y position of the current agent in the board
	 * @param j The x position of the current agent in the board
	 * @return the mergingNeighbour agent
	 */
	protected Agent crossingNeighbour ( int i, int j )
	{
		Agent agnRes = null;
		int   iNeigh = countNeighbours(i,j); 

		if ( iNeigh>0 ) {
			int iRndNei  = rnd.nextInt(iNeigh);
			int [] iaPos = getIthNeighbour(i,j,iRndNei);
		
			agnRes = agnGen[iBoard][iaPos[0]][iaPos[1]];		
 		}

		return agnRes;
	}

	//---------------------------------------------------------------------

	/** 	
	 * Returns the ith neighbor
	 * 
	 * @param i The y position of the cell
	 * @param j The x position of the cell
	 * @param iNei The neighbor searched
	 * @return The position of the neighbor
	 */
	protected int[] getIthNeighbour ( int i, int j, int iNei )
	{
		int [] iaRes = new int[2];

		iaRes[0] = -1;
		iaRes[1] = -1;

 		int iUp    = (i+1)%iMaxY;
      int iDown  = (i+iMaxY-1)%iMaxY;
      int iRight = (j+1)%iMaxX;
      int iLeft  = (j+iMaxX-1)%iMaxX;
        
		boolean bDone = false;
        
      if ( agnGen[iBoard][iUp][iLeft]!=null && !bDone ) 
			if ( iNei==0 ) {
				iaRes[0] = iUp ; iaRes[1] = iLeft;
				bDone = true;
			}
			else
				iNei--;
        
      if ( agnGen[iBoard][iUp][j]!=null && !bDone ) 
			if ( iNei==0 ) {
				iaRes[0] = iUp ; iaRes[1] = j;
				bDone = true;
			}
			else
				iNei--;

        
      if ( agnGen[iBoard][iUp][iRight]!=null && !bDone ) 
			if ( iNei==0 ) {
				iaRes[0] = iUp ; iaRes[1] = iRight;
				bDone = true;
			}
			else
				iNei--;

        
      if ( agnGen[iBoard][i][iLeft]!=null && !bDone ) 
			if ( iNei==0 ) {
				iaRes[0] = i ; iaRes[1] = iLeft;
				bDone = true;
			}
			else
				iNei--;

        
      if ( agnGen[iBoard][i][iRight]!=null && !bDone ) 
			if ( iNei==0 ) {
				iaRes[0] = i ; iaRes[1] = iRight;
				bDone = true;
			}
			else
				iNei--;

        
      if ( agnGen[iBoard][iDown][iLeft]!=null && !bDone ) 
			if ( iNei==0 ) {
				iaRes[0] = iDown ; iaRes[1] = iLeft;
				bDone = true;
			}
			else
				iNei--;

        
      if ( agnGen[iBoard][iDown][j]!=null && !bDone ) 
			if ( iNei==0 ) {
				iaRes[0] = iDown ; iaRes[1] = j;
				bDone = true;
			}
			else
				iNei--;

        
      if ( agnGen[iBoard][iDown][iRight]!=null && !bDone ) 
			if ( iNei==0 ) {
				iaRes[0] = iDown ; iaRes[1] = iRight;
				bDone = true;
			}
			else
				iNei--;

        

		return iaRes;
	}


	//---------------------------------------------------------------------
    
	/** 
	 * Implements the survival phase
	 * 
	 * @throws Exception could not clone an agent
	 */
	protected void survival()
	throws Exception
	{
		int   iNei;
		int   iNextBoard = (iBoard+1)%2;
		float fMean;
		float fStdDev;
		float fThreshold;

   	for (int i=0 ; i<iMaxY ; i++ ) 
   		for (int  j=0 ; j<iMaxX ; j++ ) 
				if ( agnGen[iBoard][i][j]!=null ) {
					iNei = countNeighbours(i,j);
					if ( iNei<=1 ) {
						// Solitude
				   	if (rnd.nextFloat()<(1.0-agnPer[i][j].getScaledAccuracy())){
							agnGen[iNextBoard][i][j] = null;
						}
						else  
							agnGen[iNextBoard][i][j] = agnGen[iBoard][i][j];
					}
					else if ( iNei>6 ) {
								// Crowded
								int [] iaPos = bestNeighbour(i,j);

								agnGen[iNextBoard][i][j] = (Agent)
										agnGen[iBoard][iaPos[0]][iaPos[1]].clone();
						  }
						  else {
							 // Steady
						 	 fMean   = neighbourMean(i,j);
						 	 fStdDev = neighbourStdDev(i,j,fMean);
							 float fKThres = fMean+galeCnf.getKThreshold()*fStdDev;
							 if ( agnPer[i][j].getScaledAccuracy()<fKThres ) {
								agnGen[iNextBoard][i][j] = null;
							 }
							 else 
								agnGen[iNextBoard][i][j] = agnGen[iBoard][i][j];
						  }
				}
				else {
					agnGen[iNextBoard][i][j] = null;
				}

		// Switching the board
		iBoard = iNextBoard;

		// Update performance references
		updatePerformance();
	}

	//---------------------------------------------------------------------

	/** 
	 * Updates the performance references.
	 *
	 * @param i The y position
	 * @param j The x position
	 * @return The number of neighbors 
	 */
	protected void updatePerformance ()
	{
		// Updating performance references
   	for (int i=0 ; i<iMaxY ; i++ ) 
   		for (int  j=0 ; j<iMaxX ; j++ ) 
				if ( agnGen[iBoard][i][j]!=null ) 
					agnPer[i][j] = agnGen[iBoard][i][j].getPerformance();
				else
					agnPer[i][j] = null;
	}

	//---------------------------------------------------------------------

	/** 
	 * Count the number of neighbors for a given cell position. 
	 *
	 * @param i The y position
	 * @param j The x position
	 * @return The number of neighbors 
	 */
	protected int countNeighbours ( int i, int j )
	{
		int iUp    = (i+1)%iMaxY;
      int iDown  = (i+iMaxY-1)%iMaxY;
      int iRight = (j+1)%iMaxX;
      int iLeft  = (j+iMaxX-1)%iMaxX;
        
      int iRes = 0;
        
      if ( agnGen[iBoard][iUp][iLeft]!=null ) 
			iRes++;
        
      if ( agnGen[iBoard][iUp][j]!=null ) 
			iRes++;
        
      if ( agnGen[iBoard][iUp][iRight]!=null ) 
			iRes++;
        
      if ( agnGen[iBoard][i][iLeft]!=null ) 
			iRes++;
        
      if ( agnGen[iBoard][i][iRight]!=null ) 
			iRes++;
        
      if ( agnGen[iBoard][iDown][iLeft]!=null ) 
			iRes++;
        
      if ( agnGen[iBoard][iDown][j]!=null ) 
			iRes++;
        
      if ( agnGen[iBoard][iDown][iRight]!=null ) 
			iRes++;
        
      return iRes;
	}
	
	//---------------------------------------------------------------------
    
	/** 
	 * Returns the position of the best neighbor for a given cell
	 *
	 * @param i The y position
	 * @param j The x position
	 * @return the best neighbor position
	 */
	protected int [] bestNeighbour( int i, int j )
   {
      int iUp    = (i+1)%iMaxY;
      int iDown  = (i+iMaxY-1)%iMaxY;
      int iRight = (j+1)%iMaxX;
      int iLeft  = (j+iMaxX-1)%iMaxX;
        
      float fBest = -100;
      int [] iaRes = new int[2];                
		
		iaRes[0] = -1;
		iaRes[1] = -1;

		if ( agnPer[iUp][iLeft]!=null )
      	if ( fBest<agnPer[iUp][iLeft].getScaledAccuracy() ) {
            fBest = agnPer[iUp][iLeft].getScaledAccuracy();
            iaRes[0] = iUp;
            iaRes[1] = iLeft;
			}
		if ( agnPer[iUp][j]!=null )
      	if ( fBest<agnPer[iUp][j].getScaledAccuracy() ) {
            fBest = agnPer[iUp][j].getScaledAccuracy();
            iaRes[0] = iUp;
            iaRes[1] = j;
      }
		if ( agnPer[iUp][iRight]!=null )
      	if ( fBest<agnPer[iUp][iRight].getScaledAccuracy() ) {
            fBest = agnPer[iUp][iRight].getScaledAccuracy();
            iaRes[0] = iUp;
            iaRes[1] = iRight;
      }
		if ( agnPer[i][iLeft]!=null )
      	if ( fBest<agnPer[i][iLeft].getScaledAccuracy() ) {
            fBest = agnPer[i][iLeft].getScaledAccuracy();
            iaRes[0] = i;
            iaRes[1] = iLeft;
      }
		if ( agnPer[i][iRight]!=null )
      	if ( fBest<agnPer[i][iRight].getScaledAccuracy() ) {
            fBest = agnPer[i][iRight].getScaledAccuracy();
            iaRes[0] = i;
            iaRes[1] = iRight;
      }
		if ( agnPer[iDown][iLeft]!=null )
      	if ( fBest<agnPer[iDown][iLeft].getScaledAccuracy() ) {
            fBest = agnPer[iDown][iLeft].getScaledAccuracy();
            iaRes[0] = iDown;
            iaRes[1] = iLeft;
      }
		if ( agnPer[iDown][j]!=null )
      	if ( fBest<agnPer[iDown][j].getScaledAccuracy() ) {
            fBest = agnPer[iDown][j].getScaledAccuracy();
            iaRes[0] = iDown;
            iaRes[1] = j;
      }
		if ( agnPer[iDown][iRight]!=null )
      	if ( fBest<agnPer[iDown][iRight].getScaledAccuracy() ) {
            fBest = agnPer[iDown][iRight].getScaledAccuracy();
            iaRes[0] = iDown;
            iaRes[1] = iRight;
      }
        
      return iaRes;
    }

	//---------------------------------------------------------------------
    
	/** 
	 * Computes the neighbor mean eval of the neighborhood of the given cell .
	 *
	 * @param i The y position
	 * @param j The x position
	 * @return the mean scaled accuracy of the neighborhood
	 */
   protected float neighbourMean ( int i, int j )
   {
    	int iUp    = (i+1)%iMaxY;
      int iDown  = (i+iMaxY-1)%iMaxY;
      int iRight = (j+1)%iMaxX;
      int iLeft  = (j+iMaxX-1)%iMaxX;
        
		int   iNei = 0;
      float fAcc = 0;

		if ( agnGen[iBoard][iUp][iLeft]!=null ) {
			fAcc += agnPer[iUp][iLeft].getScaledAccuracy();
			iNei ++;
		}

		if ( agnGen[iBoard][iUp][j]!=null ) {
			fAcc += agnPer[iUp][j].getScaledAccuracy();
			iNei ++;
		}

		if ( agnGen[iBoard][iUp][iRight]!=null ) {
			fAcc += agnPer[iUp][iRight].getScaledAccuracy();
			iNei ++;
		}

		if ( agnGen[iBoard][i][iLeft]!=null ) {
			fAcc += agnPer[i][iLeft].getScaledAccuracy();
			iNei ++;
		}

		if ( agnGen[iBoard][i][j]!=null ) {
			fAcc += agnPer[i][j].getScaledAccuracy();
			iNei ++;
		}

		if ( agnGen[iBoard][i][iRight]!=null ) {
			fAcc += agnPer[i][iRight].getScaledAccuracy();
			iNei ++;
		}

		if ( agnGen[iBoard][iDown][iLeft]!=null ) {
			fAcc += agnPer[iDown][iLeft].getScaledAccuracy();
			iNei ++;
		}

		if ( agnGen[iBoard][iDown][j]!=null ) {
			fAcc += agnPer[iDown][j].getScaledAccuracy();
			iNei ++;
		}

		if ( agnGen[iBoard][iDown][iRight]!=null ) {
			fAcc += agnPer[iDown][iRight].getScaledAccuracy();
			iNei ++;
		}
                   
      return fAcc/iNei;
   }
    
	//---------------------------------------------------------------------
    
	/** 
	 * Computes the standard deviation of the neighborhood scaled accuracy
	 * of the given cell .
	 *
	 * @param i The y position
	 * @param j The x position
	 * @param fMean the mean value
	 * @return the standard deviation of the accuracy in the neighborhood
	 */
   protected float neighbourStdDev ( int i, int j, float fMean )
   {
		int iUp    = (i+1)%iMaxY;
      int iDown  = (i+iMaxY-1)%iMaxY;
      int iRight = (j+1)%iMaxX;
      int iLeft  = (j+iMaxX-1)%iMaxX;
        
		int   iNei = 0;
      float fAcc = 0;
      float fTmp = 0;
		

		if ( agnGen[iBoard][iUp][iLeft]!=null ) {
			fTmp += agnPer[iUp][iLeft].getScaledAccuracy();
			fAcc += fTmp*fTmp;
			iNei ++;
		}

		if ( agnGen[iBoard][iUp][j]!=null ) {
			fTmp += agnPer[iUp][j].getScaledAccuracy();
			fAcc += fTmp*fTmp;
			iNei ++;
		}

		if ( agnGen[iBoard][iUp][iRight]!=null ) {
			fTmp += agnPer[iUp][iRight].getScaledAccuracy();
			fAcc += fTmp*fTmp;
			iNei ++;
		}

		if ( agnGen[iBoard][i][iLeft]!=null ) {
			fTmp += agnPer[i][iLeft].getScaledAccuracy();
			fAcc += fTmp*fTmp;
			iNei ++;
		}

		if ( agnGen[iBoard][i][j]!=null ) {
			fTmp += agnPer[i][j].getScaledAccuracy();
			fAcc += fTmp*fTmp;
			iNei ++;
		}

		if ( agnGen[iBoard][i][iRight]!=null ) {
			fTmp += agnPer[i][iRight].getScaledAccuracy();
			fAcc += fTmp*fTmp;
			iNei ++;
		}

		if ( agnGen[iBoard][iDown][iLeft]!=null ) {
			fTmp += agnPer[iDown][iLeft].getScaledAccuracy();
			fAcc += fTmp*fTmp;
			iNei ++;
		}

		if ( agnGen[iBoard][iDown][j]!=null ) {
			fTmp += agnPer[iDown][j].getScaledAccuracy();
			iNei ++;
		}

		if ( agnGen[iBoard][iDown][iRight]!=null ) {
			fTmp += agnPer[iDown][iRight].getScaledAccuracy();
			fAcc += fTmp*fTmp;
			iNei ++;
		}
                   
      return (float)Math.sqrt(fAcc/iNei);
	}

	//---------------------------------------------------------------------
    
	/** 
	 * Initializes the statistics information.
	 */
	protected void initializeStatistics()
	{
	}

	//---------------------------------------------------------------------
    
	/** 
	 * Compute the statistics of the board.
	 * 
	 * @param iIter current board iteration
	 * @param iIter maximum number of iteration
	 * @throws Exception Not valid mapping strategy
	 */
	protected void computeStatistics ( int iIter, int iMax )
	throws Exception
	{
		traceSummary(iIter);
		if ( bDump ) {
			traceAccuracy();
			traceComplexity();
		}
	}

	//---------------------------------------------------------------------
    
	/** 
	 * Summaries the 
	 */
	protected void traceSummary ( int iIter )
	throws Exception
	{
		float fMinCmplx  = Float.MAX_VALUE;
		float fMeanCmplx = 0;
		float fMaxCmplx  = -Float.MIN_VALUE;

		float fMinEval  = Float.MAX_VALUE;
		float fMeanEval = 0;
		float fMaxEval  = -Float.MIN_VALUE;

		float fTmp  = 0;
		int   iAgns = 0;

		for ( int i=0 ; i<iMaxY ; i++ ) {
			for ( int j=0 ; j<iMaxX ; j++ )
				if ( baAllIns[i][j] && agnGen[iBoard][i][j]!=null ) {
					iAgns++;
					// Accuracy
					fTmp = agnPer[i][j].getAccuracy();
					if ( fTmp>fMaxEval ) fMaxEval = fTmp;
					if ( fTmp<fMinEval ) fMinEval = fTmp;
					fMeanEval += fTmp;
					// Complexity
					fTmp = agnGen[iBoard][i][j].complexity();
					if ( fTmp>fMaxCmplx ) fMaxCmplx = fTmp;
					if ( fTmp<fMinCmplx ) fMinCmplx = fTmp;
					fMeanCmplx += fTmp;
				}
		}	

		fMeanEval /= iAgns;
		fMeanCmplx /= iAgns;

		ioFac.push("traceEval", iIter+"\t"+fMeanEval+"\t"+
		                        fMaxEval+"\t"+fMinEval);
		ioFac.push("traceCmplx", iIter+"\t"+fMeanCmplx+"\t"+
		                         fMaxCmplx+"\t"+fMinCmplx);
	}
	
	//---------------------------------------------------------------------
    
	/** 
	 * Traces the accuracy board.
	 */
	protected void traceAccuracy ()
	throws Exception
	{
		String sRes = null;

		for ( int i=0 ; i<iMaxY ; i++ ) {
			sRes = "";
			for ( int j=0 ; j<iMaxX ; j++ )
				if ( agnGen[iBoard][i][j]==null )
					sRes += ".\t";
				else
					sRes += agnPer[i][j].getAccuracy()+"\t";
			ioFac.push("eval", sRes);
		}	
		ioFac.push("eval", "--------------------------------------------------");
	}
	
	//---------------------------------------------------------------------
    
	/** 
	 * Traces the complexity board.
	 */
	protected void traceComplexity ()
	throws Exception
	{
		String sRes = null;

		for ( int i=0 ; i<iMaxY ; i++ ) {
			sRes = "";
			for ( int j=0 ; j<iMaxX ; j++ )
				if ( agnGen[iBoard][i][j]==null )
					sRes += ".\t";
				else
					sRes += agnGen[iBoard][i][j].complexity()+"\t";
			ioFac.push("cmplx", sRes);
		}	
		ioFac.push("cmplx", "--------------------------------------------------");
	}
	
	//---------------------------------------------------------------------
    
	/** 
	 * Arranges the data mapping over the board.
	 *
	 * @param os The set of available instances
	 * @throws Exception Not valid mapping strategy
	 */
	protected void arrangeMapping ( OrderedSet os )
	throws Exception
   {
		String sStra = galeCnf.getMappingStrategy();

		if (sStra.equalsIgnoreCase("uniform"))
			uniformMapping(os);
		else if (sStra.equalsIgnoreCase("random"))
			randomMapping(os);
		else if (sStra.equalsIgnoreCase("pyramidal"))
			pyramidTailNumberMapping(os);
		else
			throw new Exception (sStra+" is not a valid mapping strategy");
   }

	//---------------------------------------------------------------------

	/**
	 * Computes the number of samples covered by the individual.
	 * 
	 * @throws Exception Problems appeared when cloning
	 */
	protected void computePerformance ()
	throws Exception
	{
      int     iBestPosY  = -1;
      int     iBestPosX  = -1;
		float   fTmp       = 0;
		float   fTmpCmplx  = 0;
		float   fBestEval  = agnBest.getPerformance().getAccuracy();
		float   fBestCmplx = agnBest.complexity();
		boolean bPrune     = galeCnf.getPrune();
		// boolean bTmp       = false;
		
   	for ( int i=0 ; i<iMaxY ; i++ ) 
   		for ( int j=0 ; j<iMaxX ; j++ ) {
				if ( agnGen[iBoard][i][j]!=null ) {
					agnGen[iBoard][i][j].resetPerformance();
					//---------- Agent performance computation -------------------
					for ( int k=0, iMax=osaMapping[i][j].getSize() ; k<iMax ; k++ ){
					  Instance ins = osaMapping[i][j].getInstance(k);
					  agnGen[iBoard][i][j].classify(ins);
					}
					//------------- Pruning unused things ------------------------
					if ( bPrune ) {
				 		agnGen[iBoard][i][j] = agnGen[iBoard][i][j].prune();
				 		if (agnGen[iBoard][i][j]==null)
				 			agnPer[i][j]=null;
				 		else
				 			agnPer[i][j]=agnGen[iBoard][i][j].getPerformance();
				 	}
					//-------------- Updating the best information ----------------
					if ( agnGen[iBoard][i][j]!=null && baAllIns[i][j] ) {
						fTmp = agnPer[i][j].getAccuracy();
      				fTmpCmplx = agnGen[iBoard][i][j].complexity();
						//bTmp = agnPer[i][j].equals(agnGen[iBoard][i][j].getPerformance());
						//if ( (fBestEval<fTmp) || (fBestEval==fTmp && bTmp && fTmpCmplx<fBestCmplx) ) {
						if ( (fBestEval<fTmp) || (fBestEval==fTmp && fTmpCmplx<fBestCmplx) ) {
      					fBestEval  = fTmp;
      					fBestCmplx = fTmpCmplx;
         	   		iBestPosY  = i;
         	   		iBestPosX  = j;
	         		}
					}

	         }
			}
		if ( iBestPosX!=-1 && iBestPosY!=-1 && agnGen[iBoard][iBestPosY][iBestPosX]!=null )
			agnBest = (Agent)agnGen[iBoard][iBestPosY][iBestPosX].clone();
   }


	//---------------------------------------------------------------------

	/**
	 * Arranges the data mapping over the board uniformly
	 *
	 * @param os The set of available instances
    */
	protected void uniformMapping ( OrderedSet os )
   {
   	for ( int i=0 ; i<iMaxY ; i++ )
   		for ( int j=0 ; j<iMaxX ; j++ ) {
				osaMapping[i][j] = os;
				baAllIns[i][j]   = true;
		 	}
   }

	//---------------------------------------------------------------------

	/**
	 * Arranges the data mapping over the board with random sampling
	 *
	 * @param os The set of available instances
    */
	protected void randomMapping ( OrderedSet os )
   {
      RandomSamplingDataView rs = new RandomSamplingDataView(os,galeCnf.getInstancesProportion(),rnd);

   	for ( int i=0 ; i<iMaxY ; i++ )
   		for ( int j=0 ; j<iMaxX ; j++ ) {
				osaMapping[i][j] = rs;
				baAllIns[i][j]   = true;
		 	}
   }

   //---------------------------------------------------------------------

	/** 
	 * Arranges the data mapping over the board 
	 *
	 * @param os The set of available instances
	 */
	protected void pyramidTailNumberMapping ( OrderedSet os )
   {
		int ismSize = os.getSize();
		int iTmp    = -1;
		int iEnd    = -1;

		int iFirstSliceX  = iMaxX/3;
		int iFirstSliceY  = iMaxY/3;
		int iSecondSliceX = 2*iFirstSliceX+(iMaxX-3*iFirstSliceX);
		int iSecondSliceY = 2*iFirstSliceY+(iMaxY-3*iFirstSliceY);
	
      float fXStep = 1.0f/((float)iFirstSliceX+1);
      float fYStep = 1.0f/((float)iFirstSliceY+1);

   	for ( int i=0 ; i<iMaxY ; i++ ) {
   		for ( int j=0 ; j<iMaxX ; j++ ) {
				baAllIns[i][j]=false;
				if ( i<iFirstSliceY ) {
					if ( j<iFirstSliceX ) {
					    if ( i<j ) 	
							iEnd = (int)(ismSize*fYStep*(i+1));
						 else
							iEnd = (int)(ismSize*fXStep*(j+1));
					}
					else if ( j<iSecondSliceX ) {
						iEnd = (int)(ismSize*fYStep*(i+1));
					}
					else {
						if ( i<(iMaxX-j-1) ) 
							iEnd = (int)(ismSize*fYStep*(i+1));
						else {
							iTmp = iMaxX-j;
							iEnd = (int)(ismSize*fXStep*iTmp);
						}
					}
				}
				else if ( i<iSecondSliceY ) {
					if ( j<iFirstSliceX ) {
						iEnd = (int)(ismSize*fXStep*(j+1));
					}
					else if ( j<iSecondSliceX ) {
						iEnd = ismSize-1;	
						if ( i==iFirstSliceY      || j==iFirstSliceX ||
						     (i+1)==iSecondSliceY || (j+1)==iSecondSliceX )
							baAllIns[i][j]=false;
						else
							baAllIns[i][j]=true;
					}
					else {
						iTmp = iMaxX-j;
						iEnd = (int)(ismSize*fXStep*iTmp);
					}
				}
				else {
					if ( j<iFirstSliceX ) {
						if ( (iMaxY-i-1)<j ) {
							iTmp = iMaxY-i;
							iEnd = (int)(ismSize*fYStep*iTmp);
						}
						else
							iEnd = (int)(ismSize*fXStep*(j+1));
					}
					else if ( j<iSecondSliceX ) {
						iTmp = iMaxY-i;
						iEnd = (int)(ismSize*fYStep*iTmp);
					}
					else {
						if ( i>j ) {
							iTmp = iMaxY-i;
							iEnd = (int)(ismSize*fYStep*iTmp);
						}
						else {
							iTmp = iMaxX-j;
							iEnd = (int)(ismSize*fXStep*iTmp);
						}
					}
				}
				//System.out.println(j+" "+i+" "+iEnd);
				osaMapping[i][j] = new DataView(os,0,iEnd);
		 	}
		}
   }

	//-----------------------------------------------------------------------

}
